# -*- coding: utf-8 -*-
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.routing import Route
import starlette_werkzeug_debugger


async def raise_error(request):
	local_var = 3
	raise RuntimeError("Raised error")


middleware = [
	Middleware(starlette_werkzeug_debugger.WerkzeugDebugMiddleware, evalex=True)
]


app = Starlette(debug=True, middleware=middleware, routes=[
	Route('/', raise_error),
])



# -*- coding: utf-8 -*-
import re
import typing as t
from io import BytesIO

from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp, Receive, Scope, Send
from werkzeug.debug import DebugTraceback, DebuggedApplication
from werkzeug.wrappers import Response as WerkzeugResponse


class WerkzeugDebugMiddleware(DebuggedApplication):
	DEBUGGER_REQUEST_RX = re.compile(r'^(?:.*&)?__debugger__=yes(?:&.*)?$')

	def __init__(
		self, app: ASGIApp,
		**kwargs
	):
		super().__init__(None, **kwargs)
		self.app = app

	def get_wsgi_environ(self, request: Request) -> dict:
		return {
			'REQUEST_METHOD': request.method,
			'HTTP_COOKIE': request.headers.get('cookie'),
			'request': request,
		}

	async def debugger_response(self, scope: Scope, receive: Receive, send: Send) -> bool:
		request = Request(scope, receive, send)
		response = None

		request.environ = self.get_wsgi_environ(request)
		request.args = request.query_params
		request.is_secure = False

		cmd = request.query_params.get("cmd")
		arg = request.query_params.get("f")
		secret = request.query_params.get("s")
		frm = request.query_params.get("frm")
		if frm is not None:
			frm = int(frm)
		frame = self.frames.get(frm)  # type: ignore
		if cmd == "resource" and arg:
			response = self.get_resource(request, arg)  # type: ignore
		elif cmd == "pinauth" and secret == self.secret:
			response = self.pin_auth(request)  # type: ignore
		elif cmd == "printpin" and secret == self.secret:
			response = self.log_pin_request()  # type: ignore
		elif (
			self.evalex
			and cmd is not None
			and frame is not None
			and self.secret == secret
			and self.check_pin_trust(self.get_wsgi_environ(request))
		):
			response = self.execute_command(request, cmd, frame)  # type: ignore

		if isinstance(response, WerkzeugResponse):
			data = BytesIO()
			for chunk in response.get_app_iter(request.environ):
				data.write(chunk)
			response = Response(
				data.getvalue(),
				status_code=response.status_code,
				media_type=response.mimetype,
				headers=response.headers,
			)

		if response is not None:
			await response(scope, receive, send)
			return True
		return False

	async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
		if scope["type"] != 'http':
			await self.app(scope, receive, send)
			return

		query = scope['query_string'].decode('utf-8')
		if self.DEBUGGER_REQUEST_RX.match(query):
			if await self.debugger_response(scope, receive, send):
				return

		try:
			await self.app(scope, receive, send)
		except Exception as e:
			contexts: list[t.ContextManager[t.Any]] = []

			tb = DebugTraceback(e, skip=1, hide=not self.show_hidden_frames)

			for frame in tb.all_frames:
				self.frames[id(frame)] = frame
				self.frame_contexts[id(frame)] = contexts

			request = Request(scope, receive, send)
			is_trusted = bool(self.check_pin_trust(self.get_wsgi_environ(request)))
			html = tb.render_debugger_html(
				evalex=self.evalex,
				secret=self.secret,
				evalex_trusted=is_trusted,
			)
			response = Response(html, status_code=500, media_type="text/html")

			await response(scope, receive, send)



# -*- coding: utf-8 -*-
import contextlib
import json
import logging
import re

import starlette_werkzeug_debugger
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.responses import JSONResponse
from starlette.routing import Route
from starlette.testclient import TestClient


def inner_error():
	local_var = 'inner'
	raise RuntimeError("Raised error")


async def raise_error(request):
	local_var = 'outer'
	inner_error()


async def ok_response(request):
	return JSONResponse("ok")


@contextlib.asynccontextmanager
async def lifespan(app):
	# only to test scope["type"]
	yield


def build_app(**kwargs):
	middleware = [
		Middleware(starlette_werkzeug_debugger.WerkzeugDebugMiddleware, **kwargs)
	]

	return Starlette(debug=True, middleware=middleware, lifespan=lifespan, routes=[
		Route('/', raise_error),
		Route('/ok/', ok_response),
	])


def get_middleware(app):
	return app.middleware_stack.app


def test_correct_response():
	app = build_app()
	with TestClient(app) as client:
		response = client.get('/ok/')
		assert response.status_code == 200
		assert response.content == b'"ok"'


def test_error_response():
	app = build_app()
	client = TestClient(app)
	response = client.get('/')
	assert response.status_code == 500
	assert b"Werkzeug Debugger" in response.content


def test_serve_static():
	app = build_app()
	client = TestClient(app)
	client.get('/')
	response = client.get('/', params={'__debugger__': 'yes', 'cmd': 'resource', 'f': 'style.css'})
	assert response.status_code == 200
	assert response.headers['content-type'].startswith('text/css')


def test_printpin(caplog):
	caplog.set_level(logging.INFO)

	app = build_app(evalex=True, pin_security=True, pin_logging=True)
	client = TestClient(app)
	client.get('/')
	middleware = get_middleware(app)
	middleware.pin = '4852'

	# dont' print anything
	response = client.get('/', params={'__debugger__': 'yes', 'cmd': 'printpin', 's': middleware.secret + 'x'})
	assert middleware.pin not in caplog.text
	response = client.get('/', params={'__debugger__': 'yes', 'cmd': 'printpin', 's': middleware.secret})
	assert middleware.pin in caplog.text


def test_pinauth():
	app = build_app(evalex=True, pin_security=True, pin_logging=True)
	client = TestClient(app)
	client.get('/')
	middleware = get_middleware(app)
	middleware.pin = '4852'

	# wrong secret
	response = client.get('/', params={'__debugger__': 'yes', 'cmd': 'pinauth', 'pin': middleware.pin, 's': middleware.secret + 'x'})
	assert response.status_code == 500

	# wrong pin
	response = client.get('/', params={'__debugger__': 'yes', 'cmd': 'pinauth', 'pin': middleware.pin + '5', 's': middleware.secret})
	assert response.status_code == 200
	response_content = json.loads(response.content.decode('utf-8'))
	assert not response_content['auth']

	# correct pin
	response = client.get('/', params={'__debugger__': 'yes', 'cmd': 'pinauth', 'pin': middleware.pin, 's': middleware.secret})
	assert response.status_code == 200
	response_content = json.loads(response.content.decode('utf-8'))
	assert response_content['auth']
	assert middleware.pin_cookie_name in response.cookies


def test_console():
	app = build_app(evalex=True, pin_security=True, pin_logging=True)
	client = TestClient(app)
	exception_content = client.get('/').content.decode('utf-8')
	middleware = get_middleware(app)
	middleware.pin = '4852'

	# login
	response = client.get('/', params={'__debugger__': 'yes', 'cmd': 'pinauth', 'pin': middleware.pin, 's': middleware.secret})
	cookies = response.cookies

	frame_ids = re.findall(r'(?:frame-(\d+))', exception_content)

	# content from inner variable
	response = client.get('/', params={'__debugger__': 'yes', 'cmd': 'local_var', 'frm': frame_ids[-1], 's': middleware.secret})
	assert b'inner' in response.content

	# content from outer variable
	response = client.get('/', params={'__debugger__': 'yes', 'cmd': 'local_var', 'frm': frame_ids[-2], 's': middleware.secret})
	assert b'outer' in response.content



# -*- coding: utf-8 -*-
from fastapi import FastAPI
import starlette_werkzeug_debugger


app = FastAPI()


app.add_middleware(starlette_werkzeug_debugger.WerkzeugDebugMiddleware, evalex=True)


@app.get("/")
async def raise_error():
	local_var = 3
	raise RuntimeError("Raised error")



# -*- coding: utf-8 -*-
import json
from pathlib import Path
from typing import Any, Dict, Optional

import starlette_werkzeug_debugger
from fastapi import FastAPI
from fastapi.encoders import jsonable_encoder
from fastapi.staticfiles import StaticFiles
from starlette.responses import HTMLResponse


app = FastAPI(docs_url=None)


app.add_middleware(starlette_werkzeug_debugger.WerkzeugDebugMiddleware, evalex=True)
app.mount('/static/', StaticFiles(directory=Path(__file__).parent / 'static'), name="static")


swagger_ui_default_parameters = {
	"dom_id": "#swagger-ui",
	"layout": "BaseLayout",
	"deepLinking": True,
	"showExtensions": True,
	"showCommonExtensions": True,
}


def get_swagger_ui_html(
	*,
	openapi_url: str,
	title: str,
	swagger_js_url: str = "https://cdn.jsdelivr.net/npm/swagger-ui-dist@4/swagger-ui-bundle.js",
	swagger_css_url: str = "https://cdn.jsdelivr.net/npm/swagger-ui-dist@4/swagger-ui.css",
	swagger_favicon_url: str = "https://fastapi.tiangolo.com/img/favicon.png",
	oauth2_redirect_url: Optional[str] = None,
	init_oauth: Optional[Dict[str, Any]] = None,
	swagger_ui_parameters: Optional[Dict[str, Any]] = None,
) -> HTMLResponse:
	current_swagger_ui_parameters = swagger_ui_default_parameters.copy()
	if swagger_ui_parameters:
		current_swagger_ui_parameters.update(swagger_ui_parameters)

	swagger_js_extra_url = "/static/swagger_extra.js"
	swagger_css_extra_url = "/static/swagger_extra.css"

	html = f"""
	<!DOCTYPE html>
	<html>
	<head>
	<link type="text/css" rel="stylesheet" href="{swagger_css_url}">
	<link type="text/css" rel="stylesheet" href="{swagger_css_extra_url}">
	<link rel="shortcut icon" href="{swagger_favicon_url}">
	<title>{title}</title>
	</head>
	<body>
	<div id="swagger-ui">
	</div>
	<script src="{swagger_js_url}"></script>
	<script src="{swagger_js_extra_url}"></script>
	<!-- `SwaggerUIBundle` is now available on the page -->
	<script>
	const ui = SwaggerUIBundle({{
		url: '{openapi_url}',
		responseInterceptor: window.responseInterceptor,
	"""

	for key, value in current_swagger_ui_parameters.items():
		html += f"{json.dumps(key)}: {json.dumps(jsonable_encoder(value))},\n"

	if oauth2_redirect_url:
		html += f"oauth2RedirectUrl: window.location.origin + '{oauth2_redirect_url}',"

	html += """
	presets: [
		SwaggerUIBundle.presets.apis,
		SwaggerUIBundle.SwaggerUIStandalonePreset
		],
	})"""

	if init_oauth:
		html += f"""
		ui.initOAuth({json.dumps(jsonable_encoder(init_oauth))})
		"""

	html += """
	</script>
	</body>
	</html>
	"""
	return HTMLResponse(html)


@app.get("/docs", include_in_schema=False)
async def custom_swagger_ui_html():
	return get_swagger_ui_html(
		openapi_url=app.openapi_url,
		title=app.title + " - Swagger UI",
	)


@app.get("/")
async def raise_error():
	local_var = 3
	raise RuntimeError("Raised error")
